#! /usr/bin/env python3
#  
#  This code allow for the discretization                              
#  of a given Grain Size Distribution, following the approach            
#  proposed by Prof. Gary Parker in his e-book (Ch 2), available at:     
#  http://hydrolab.illinois.edu/people/parkerg/morphodynamics_e-book.htm              
#                                                                       
#  This code is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.
#  
#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR 
#  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
#  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
#  OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  This code has been developed by Seline Frei, Matteo Facchini and Davide Vanzo                                                                                   #
#  mailto: frei@vaw.baug.ethz.ch, mat.facchini@gmail.com and vanzo@vaw.baug.ethz.ch

#  Last revision: June 2023   


import numpy as np
import pandas as pd                 ## panda dataframe
import csv
import os
import matplotlib.pyplot as plt     ## Bibliothek fuer Matlab-aehnliches Plotten
import matplotlib.ticker as mtick   ## zur Modifizierung der Achsen-Teilstriche
from matplotlib.patches import Rectangle ## Darstellung der Rectangles in matplotlib
import matplotlib.patches as mpatches ## Darstellung der Legende (Rectangles)
import math 
import iteround                     ## fractions runden und Summe (100%, 1) exakt erhalten


def save_figure(fig,filename):
    #fig.savefig(filename+'.pdf', dpi = 300, format='pdf')
    fig.savefig(filename+'.png', dpi = 300, format='png')

def print_welcome():
    print('# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #')
    print('#                                                                       #')
    print('# This script allow for the discretization                              #')
    print('# of a given Grain Size Distribution, following the approach            #')
    print('# proposed by Prof. Gary Parker in his e-book (Ch 2), available at:     #')
    print('# http://hydrolab.illinois.edu/people/parkerg/morphodynamics_e-book.htm #')
    print('#                                                                       #')
    print('# This code has been developed by S. Frei, M. Facchini and D. Vanzo     #')
    print('#                                                                       #')
    print('# mailto: selifrei@vaw.baug.ethz.ch, mat.facchini@gmail.com,            #') 
    print('#         vanzo@vaw.baug.ethz.ch                                        #')
    print('#                                                                       #')
    print('# Neither we nor any university/institution in which we are employed    #')
    print('# accepts responsibility or liability for its use by third parties.     #')
    print('#                                                                       #')
    print('# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #')

def readGSD(pathin,fname):
    ''' Read the input CSV file with the grain size distribution in the form diameter [mm], cum fraction [-] '''
    with open(pathin+fname+'.csv') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        header = next(csv_reader)
        d = []
        f = []
        for row in csv_reader:
            d.append(row[0])
            f.append(row[1])
    d = np.array(d,float)
    f = np.array(f,float)
    # correct the distribution if f is not exactly from 0 to 1 (to be revised)
    dk,fk = correctGSD(d,f)
    psik = np.log2(dk)
    # classes (Ch 2 slide 8, Parker ebook)
    psikc = 0.5*(psik[:-1]+psik[1:])
    dkc = 2**psikc
    fkc = fk[1:]-fk[:-1]
    # dictionaries
    gsd  = {'diameters':dk,'fractions':fk,'psi':psik}
    gdsc = {'diameters':dkc,'fractions':fkc,'psi':psikc}

    return gsd,gdsc

def correctGSD(di,fi):
    '''correct the distribution if f is not exactly from 0 to 1 (to be revised)'''
    psii = np.log2(di)
    if fi[0] != 0.0:
        psi0 = interpolateD(di,fi,0.0)
        #add interpolated values to data (position 0)
        psii = np.insert(psii,0,psi0)
        di   = np.insert(di,0,2**psi0)
        fi   = np.insert(fi,0,0.0)
    if fi[-1] != 1.0:
        psiN = interpolateD(di,fi,1.0)
        #add interpolated values to data (last position)
        np.append(psii,psiN)
        di = np.append(di,2**psiN)
        fi = np.append(fi, 1.0)
    return di, fi

def interpolateD(di,fi,intF):
    '''manually defined interpolator'''
    psi = np.log2(di)
    # interpolation boundaries ID
    if intF == 0.0:
        ip1 = 1
    elif intF == 1.0:
        ip1 = -1
    else:
        ip1 = np.where(fi>intF)[0][0]
    im1 = ip1-1
    # coefficients
    df   = fi[ip1] - fi[im1] # delta f
    dpsi = psi[ip1] - psi[im1] # delta psi
    m = df/dpsi # slope
    q = fi[im1]-m*psi[im1] # y-intercept
    # calculate
    intPsi = (intF-q)/m # f(psi)=m*psi+q  --> 0=m*psi+q --> psi=(0-q)/m or 1=m*psi+q --> psi=(1-q)/m
    intDi  = 2**intPsi
    return intPsi

def interpolateF(di,fi,intD):
    intDm  = intD
    psii   = np.log2(di)
    intPsi = np.log2(intDm)
    # check intD
    if (intDm-di[0])<1e-14 or (intDm-di[-1])>1e-14:
        print('INTERPOLATION ERROR: diameter out of GSD bounds')
        exit()
    # interpolation boundaries ID
    if intD == di[0]:
        ip1 = 1
    elif intD == di[-1]:
        ip1 = -1
    else:
        ip1 = np.where(di>intDm)[0][0]
    im1 = ip1-1
    # coefficients
    df   = fi[ip1] - fi[im1]
    dpsi = psii[ip1] - psii[im1]
    m = df/dpsi
    q = fi[im1]-m*psii[im1]
    # calculate
    intF  = m*intPsi+q 
    intDi = 2**intPsi
    return intF

def d_x(x,gsd):
    # correction
    di = gsd['diameters']
    fi = gsd['fractions']
    print(2**interpolateD(di,fi,x)) 

def dmsg(gsd):
    dic  = gsd['diameters']
    fic  = gsd['fractions']
    psic = np.log2(dic)
    # dm
    dm = np.dot(dic,fic)
    # dg
    psig = np.dot(psic,fic)
    #dg = 2**psig
    # sigmag
    psiMpsig = np.array([(psii-psig)**2 for psii in psic],float)
    sigma2   = np.dot(psiMpsig,fic)
    sigma    = np.sqrt(sigma2)
    sigmag   = 2**sigma
    
    return dm,sigmag

def d_90(gsd):
    dk = gsd['diameters']
    fk = gsd['fractions']
    # calculate d90: 90% finer
    psii90 = interpolateD(dk,fk,0.9)
    d90 = 2**psii90
    return d90



def discGSDnD(gsd,N):
    # original distribution
    d0   = gsd['diameters']
    psi0 = np.log2(d0)
    f0   = gsd['fractions']
    # edges
    dpsi      = (psi0[-1]-psi0[0])/N # increment
    psi_edges = [psi0[0]] # x edge boundaries
    f_edges   = [f0[0]] # y edge boundaries
    for i in range(N-1):
        psi_edge = psi_edges[i]+dpsi
        d_edge   = 2**psi_edge
        f_edge   = interpolateF(d0,f0,d_edge)
        psi_edges.append(psi_edge)
        f_edges.append(f_edge)
    # check if it sums to the last diameter
    dpsi_end = abs((psi_edge+dpsi)-psi0[-1])
    if dpsi_end<1e-14:
        psi_edges.append(psi_edge+dpsi)
        f_edges.append(1.0)
    d_edges = [2**psie for psie in psi_edges]
    psi_edges = np.array(psi_edges,float)
    d_edges = np.array(d_edges,float)
    f_edges = np.array(f_edges,float)
    # classes 
    psi_classes = (psi_edges[:-1]+psi_edges[1:])/2
    d_classes = [2**psic for psic in psi_classes]
    d_classes = np.array(d_classes,float)
    f_classes = f_edges[1:]-f_edges[:-1]
    # output
    GSDinterpolated = {'diameters':d_edges,'fractions':f_edges}
    GSDclasses = {'diameters':d_classes,'fractions':f_classes}
    return GSDinterpolated, GSDclasses



def writeGSDd(pathout, fname, GSDint):
    # dimensions
    Ne = len(GSDint['diameters'])
    Nc = Ne-1
    print('save distribution (%g gc) to file' %Nc)
    # diameters 
    dsk_int = GSDint['diameters']
    fsk_int = GSDint['fractions']
    if not os.path.exists(pathout):
        os.makedirs(pathout)
    # distribution
    fname_distributions  = 'GSD_distribution_%dgc' % Nc
    fname_distributions += fname+'.dat'
    if os.path.isfile(pathout+fname_distributions):
        overwrite = input('File %s already exists. Overwrite it? (Y/N) ' % fname_distributions).lower()
    else:
        overwrite = 'y'
    if overwrite == 'y':
        with open(pathout+fname_distributions, 'w') as outfile1:
            outfile1.write('# approximation to %.14g grain size classes\n# Approximated GSD \n' % Nc)
            outfile1.write('diameter [mm]    fraction [-]\n')
            for i in range(Ne):
                outfile1.write('%.0f\t%.14g\n' % (dsk_int[i],fsk_int[i]))

def writeGSDc(pathout, fname, GSDcla):
    # dimensions
    Nc = len(GSDcla['diameters'])
    print('save %g classes to file' %Nc)
    # diameters 
    dsk_cla = GSDcla['diameters']
    fsk_cla = GSDcla['fractions']
    if not os.path.exists(pathout):
        os.makedirs(pathout)
    # classes   
    fname_classes  = 'GSD_classes_%dgc' % Nc
    fname_classes += fname+'.dat'
    if os.path.isfile(pathout+fname_classes):
        overwrite = input('File %s already exists. Overwrite it? (Y/N) ' % fname_classes).lower()
    else:
        overwrite = 'y'
    if overwrite == 'y':
        with open(pathout+fname_classes, 'w') as outfile:
            outfile.write('# approximation to %.14g grain size classes, as input for BASEMENT\n' % Nc)
            outfile.write('diameter [mm]    fraction [-]\n')
            for i in range(Nc):
                outfile.write('%.0f\t%.14g\n' % (dsk_cla[i],fsk_cla[i]))

def print_stats(dm,sg):
    dm = '{0:.0f}'.format(dm)
    sg = '{0:.1f}'.format(sg)
    print('\n')
    print('Mean diameter dm: ', dm, ' mm')
    print('Geometric standard deviation sigma_g: ', sg, ' mm')

def print_d90(d90):
    d90 = '{0:.0f}'.format(d90)
    print('d90: ', d90, ' mm')    


def Abbildung(input_distribution, input_gsd, input_gsc, name_gsd, lin_log):
    
    plt.rc('savefig', bbox='tight')         ## Abbildungen randlos speichern
    # Formatierung der Teilstriche (Masse in pt)
    plt.rc('xtick', direction='out')        ## Richtung x-Teilstriche
    plt.rc('ytick', direction='out')        ## Richtung y-Teilstriche
    plt.rc('xtick.major', pad=5)           ## Abstand x-Teilstrich - Beschriftung
    plt.rc('ytick.major', pad=5)           ## Abstand y-Teilstrich - Beschriftung
    plt.rc('xtick.major', size=5)          ## Laenge der x-Haupt-Teilstriche
    plt.rc('ytick.major', size=5)          ## Laenge der y-Haupt-Teilstriche
    plt.rc('xtick.minor', size=2)          ## Laenge der x-Unter-Teilstriche
    plt.rc('ytick.minor', size=2)          ## Laenge der y-Unter-Teilstriche

    # Formatierung des Gitters
    plt.rc('grid', linewidth=0.75)           ## Linienbreite (in pt)
    plt.rc('grid', linestyle='-.')          ## Linientyp (gestrichelt)
 
    # ------------------------------------------------------------------------------
    # Korngrössenverteilung
    # ------------------------------------------------------------------------------
    # fractions in % umrechnen
    input_gsd['fractions'] = input_gsd['fractions']*100
    input_gsc['fractions'] = input_gsc['fractions']*100
    input_gsc['cumsum']    = input_gsc['cumsum']*100
    input_distribution['fractions'] = input_distribution['fractions']*100
    
    ax = plt.gca()
    fig = plt.gcf()

    #d-Linien Plotten
    for i in range (len(input_gsc)):
        X = input_gsc['diameters'][i], input_gsc['diameters'][i]
        if i==0:
            Y = [0, input_gsc['cumsum'][i]]
        else:
            Y = [input_gsc['cumsum'][i-1], input_gsc['cumsum'][i]]
        d_line, = plt.plot(X,Y, color='blue', linewidth=1, label='Mean size-class diameter')

    # Input Distribution plotten
    input_distr, = plt.plot(input_distribution['diameters'], input_distribution['fractions'], color='black',
        linestyle='-', label='Original grain size distribution', marker='x', markersize=5, linewidth=1)
     
    #Boxen plotten
    for i in range(len(input_gsc)):
        i=i+1
        ax.add_patch(Rectangle((input_gsd['diameters'][i],input_gsd['fractions'][i]),input_gsd['diameters'][i+1]-input_gsd['diameters'][i],input_gsd['fractions'][i+1]-input_gsd['fractions'][i],
            alpha=0.3,color='blue',linewidth = 1.5))

    # Beschriftung Boxen plotten
    for i in range(len(input_gsc)):
        plt.annotate(str(round(input_gsc.iloc[i,1]))+'% ', xy=(input_gsd.iloc[i+1,0], ((input_gsd.iloc[i+2,1]-input_gsd.iloc[i+1,1])/2 + input_gsd.iloc[i+1,1])-1),
                     color='blue',
                     horizontalalignment='right',
                     fontsize=7)
    
    # Legende anzeigen
    # create patch for legend
    patch = mpatches.Patch(alpha=0.3, color='blue', linewidth = 1.5, label='Grain size class')
    
    plt.legend(handles=[input_distr, patch, d_line],loc='upper center', bbox_to_anchor=(0.5, -0.3),
        fancybox=False, shadow=False, ncol=1)

    # Gitter anzeigen
    plt.grid()
    # Setze Gitter in den Hindergrund
    ax.set_axisbelow(True)

    # Zweite x-Achse erstellen
    ax2 = ax.twiny()

    # Labels erzeugen
    ax.set_xlabel('Grain size  [mm]', fontweight='bold')
    ax2.set_xlabel('Mean size-class diameter [mm]', fontsize = 7, color='blue')
    ax.set_ylabel('Percent finer [%]', fontweight='bold')

    # Groesse Plot definieren
    fig.set_size_inches(3,3)

    #logarithmische x-Achse erstellen
    ax.set_xscale('log')
    ax2.set_xscale('log')
        
    #x-Achse beschränken
    #Eingabe speichern
    xmin = float(input('x-axis minimum, grain size in mm: '))
    xmax = float(input('x-axis maximum, grain size in mm: ')) 
    # Achse beschränken  
    ax.set_xlim(xmin,xmax)
    ax2.set_xlim(xmin,xmax)

    # xticks erstellen
    # major x-tick
    xmin_pow = math.floor(math.log10(xmin))
    xmax_pow = math.ceil(math.log10(xmax))        
    x_ticks = [10**i for i in range(xmin_pow, xmax_pow+1)]
    ax.set_xticks(x_ticks)

    # xticks show normal and not scientific 10^xx
    ax.xaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: '{:.16g}'.format(y)))

    # minor xticks
    # include minor xticks
    x_minor_ticks = mtick.LogLocator(base = 10.0, subs = np.arange(1.0, 10.0) * 0.1, numticks = 10)
    ax.xaxis.set_minor_locator(x_minor_ticks)
    ax.xaxis.set_minor_formatter(mtick.NullFormatter())

    # y-Achse beschränken
    plt.ylim(0,100)
    # xticks gemäss BM-Korndurchmesser erstellen
    ax2.set_xticks( input_gsc['diameters'][:] )
    # Labels xticks BM-Korndurchmesser erstellen
    ax2.set_xticklabels(input_gsc['diameters'][:], fontsize = 7, color='dimgray', rotation=90)
    ax2.xaxis.set_major_formatter(mtick.FormatStrFormatter('%.0f'))
    # minorticks wegnehmen
    ax2.minorticks_off()
    # make xticks x-axis grey
    ax2.tick_params(axis='x', colors='blue') 

    # Achse beschränken  
    ax.set_xlim(xmin,xmax)
    ax2.set_xlim(xmin,xmax)

    #generated with BASEgsd
    plt.annotate('generated with', xy=(51, 269), xycoords='figure pixels',
                     color='grey',
                     horizontalalignment='left',
                     fontsize=6)

    plt.annotate('BASEgsd', xy=(51, 260), xycoords='figure pixels',
                     color='grey',
                     horizontalalignment='left',
                     fontsize=7)


    # Abbildung speichern
    name_plot = 'BASEgsd_' + str(len(input_gsc)) + '_GSC' + '.pdf'
    plt.savefig(name_plot)

    # Abbildung schliessen
    plt.close()
    return



if __name__ == '__main__':
    print_welcome()
    print('\n')
    print('** basic assumptions: (i) diameters are give in mm, the GSD is cumulative **')
    print('\n')
    # choice
    pathin  = './input/'
    pathout = './output/'
    fnamein = 'Distribution' #name of input file
    inGSD, inGSDc = readGSD(pathin,fnamein) #read input file and save
    print('\n')

    #calculate and print main statistics (Ch 2, slide 9 Parker ebook)
    dm,sg = dmsg(inGSDc) #calculate dm and sigma_g
    d90   = d_90(inGSD)  #calculate d90
    print('ORIGINAL distribution')
    print('Number N+1 of discrete points (with extrapolation): %i'%len(inGSD['diameters']))
    print('Number N of discrete classes: %i'%len(inGSDc['diameters']))
    print_stats(dm,sg)
    print_d90(d90)
    print('\n ************************************** \n')
    
    fname = r'_diam_log'
    Nclasses = int(input('Number of classes: '))
    interpGSD, classesGSD = discGSDnD(inGSD,Nclasses)
    

    # round values and preserve sum (=100% =1)
    classesGSD['fractions'] = iteround.saferound(classesGSD['fractions'], 2)
    classesGSD['diameters']  = np.round(classesGSD['diameters']) #round to mm
    interpGSD['fractions'] = np.append(np.array([0]), np.cumsum(classesGSD['fractions']))
    interpGSD['diameters']  = np.round(interpGSD['diameters']) #round to mm    

    # write Output files with rounded values
    writeGSDd(pathout, fname, interpGSD)
    writeGSDc(pathout, fname, classesGSD)
    # calculate dm and sg for GSD with rounded values
    dm,sg = dmsg(classesGSD)
    print('Diameter-based (log scale) discretized distribution, %g gc' %Nclasses)
    print_stats(dm,sg)
    print('\n')

    # plot results
    #preparation panada DataFrame for Plot
    df_classesGSD = pd.DataFrame(data=classesGSD)
    #calculate cumulative sum for plotting
    df_classesGSD['cumsum'] = df_classesGSD.iloc[:,1].cumsum()

    #create pandas Dataframe
    df_interpGSD = pd.DataFrame(data=interpGSD)
    #add support line for plotting (first line) and correct index
    df_interpGSD.loc[-1] = ['0', '0']
    df_interpGSD.index = df_interpGSD.index + 1  # shifting index
    df_interpGSD.sort_index(inplace=True) 
    df_interpGSD = df_interpGSD.astype(float)
    #add support line for plotting (last line)
    df_interpGSD.loc[len(df_interpGSD)] = [df_interpGSD.iloc[-1,0], 1]

    #create pandas Dataframe
    df_inGSD = pd.DataFrame(data=inGSD)

    #plot erstellen und speichern
    Abbildung(df_inGSD, df_interpGSD, df_classesGSD, fname, 1)










